<?php

/**
 * @file
 * Contains \Drupal\big_pipe\Tests\BigPipeTest.
 */

namespace Drupal\big_pipe\Tests;

use Drupal\Component\Serialization\Json;
use Drupal\Component\Utility\Html;
use Drupal\Core\Url;
use Drupal\simpletest\WebTestBase;

/**
 * Tests BigPipe's no-JS detection & response delivery (with and without JS).
 *
 * Covers:
 * - big_pipe_page_attachments()
 * - \Drupal\big_pipe\Controller\BigPipeController
 * - \Drupal\big_pipe\EventSubscriber\HtmlResponseBigPipeSubscriber
 * - \Drupal\big_pipe\Render\BigPipe
 *
 * @group big_pipe
 */
class BigPipeTest extends WebTestBase {

  const START_SIGNAL= '<script type="application/json" data-big-pipe-event="start"></script>';
  const STOP_SIGNAL= '<script type="application/json" data-big-pipe-event="stop"></script>';

  /**
   * Modules to enable.
   *
   * @var array
   */
  public static $modules = ['big_pipe', 'big_pipe_test'];

  /**
   * {@inheritdoc}
   */
  protected $dumpHeaders = TRUE;

  /**
   * {@inheritdoc}
   */
  protected function setUp() {
    parent::setUp();

    // Ignore the <meta> refresh that big_pipe.module sets. It causes a redirect
    // to a page that sets another cookie, which causes WebTestBase to lose the
    // session cookie. To avoid this problem, tests should first call
    // drupalGet() and then call checkForMetaRefresh() manually, and then reset
    // $this->maximumMetaRefreshCount and $this->metaRefreshCount.
    // @see doMetaRefresh()
    $this->maximumMetaRefreshCount = 0;
  }

  /**
   * Performs a single <meta> refresh explicitly.
   *
   * This test disables the automatic <meta> refresh checking, each time it is
   * desired that this runs, a test case must explicitly call this.
   *
   * @see setUp()
   */
  protected function performMetaRefresh() {
    $this->maximumMetaRefreshCount = 1;
    $this->checkForMetaRefresh();
    $this->maximumMetaRefreshCount = 0;
    $this->metaRefreshCount = 0;
  }

  /**
   * Tests BigPipe's no-JS detection.
   *
   * Covers:
   * - big_pipe_page_attachments()
   * - \Drupal\big_pipe\Controller\BigPipeController
   */
  public function testNoJsDetection() {
    // 1. No session (anonymous).
    $this->drupalGet(Url::fromRoute('<front>'));
    $this->assertSessionCookieExists(FALSE);
    $this->assertBigPipeNoJsCookieExists(FALSE);
    $this->assertNoRaw('<noscript><meta http-equiv="Refresh" content="0; URL=');

    // 2. Session (authenticated).
    $this->drupalLogin($this->rootUser);
    $this->assertSessionCookieExists(TRUE);
    $this->assertBigPipeNoJsCookieExists(FALSE);
    $this->assertRaw('<noscript><meta http-equiv="Refresh" content="0; URL=' . base_path() . 'big_pipe/no-js?destination=' . base_path() . 'user/1" />' . "\n" . '</noscript>');
    $this->assertBigPipeNoJsMetaRefreshRedirect();
    $this->assertBigPipeNoJsCookieExists(TRUE);

    // Close the prior connection and remove the collected state.
    $this->curlClose();
    $this->curlCookies = [];
    $this->cookies = [];

    // 3. Session (anonymous).
    $this->drupalGet(Url::fromRoute('user.login', [], ['query' => ['trigger_session' => 1]]));
    $this->drupalGet(Url::fromRoute('user.login'));
    $this->assertSessionCookieExists(TRUE);
    $this->assertBigPipeNoJsCookieExists(FALSE);
    $this->assertRaw('<noscript><meta http-equiv="Refresh" content="0; URL=' . base_path() . 'big_pipe/no-js?destination=' . base_path() . 'user/login" />' . "\n" . '</noscript>');
    $this->assertBigPipeNoJsMetaRefreshRedirect();
    $this->assertBigPipeNoJsCookieExists(TRUE);
  }

  /**
   * Tests BigPipe-delivered HTML responses when JavaScript is enabled.
   *
   * Covers:
   * - \Drupal\big_pipe\EventSubscriber\HtmlResponseBigPipeSubscriber
   * - \Drupal\big_pipe\Render\BigPipe
   * - \Drupal\big_pipe\Render\BigPipe::sendPlaceholders()
   *
   * @see \Drupal\big_pipe\Tests\BigPipePlaceholderTestCases
   */
  public function testBigPipe() {
    $this->drupalLogin($this->rootUser);
    $this->assertSessionCookieExists(TRUE);
    $this->assertBigPipeNoJsCookieExists(FALSE);

    // By not calling performMetaRefresh() here, we simulate JavaScript being
    // enabled, because as far as the BigPipe module is concerned, JavaScript is
    // enabled in the browser as long as the BigPipe no-JS cookie is *not* set.
    // @see setUp()
    // @see performMetaRefresh()

    $this->drupalGet(Url::fromRoute('big_pipe_test'));
    $this->assertBigPipeResponseHeadersPresent();

    $cases = $this->getTestCases();
    $this->assertBigPipeNoJsPlaceholders([
      $cases['edge_case__invalid_html']->bigPipeNoJsPlaceholder     => $cases['edge_case__invalid_html']->embeddedHtmlResponse,
      $cases['html_attribute_value']->bigPipeNoJsPlaceholder        => $cases['html_attribute_value']->embeddedHtmlResponse,
      $cases['html_attribute_value_subset']->bigPipeNoJsPlaceholder => $cases['html_attribute_value_subset']->embeddedHtmlResponse,
    ]);
    $this->assertBigPipePlaceholders([
      $cases['html']->bigPipePlaceholderId                             => Json::encode($cases['html']->embeddedAjaxResponseCommands),
      $cases['edge_case__html_non_lazy_builder']->bigPipePlaceholderId => Json::encode($cases['edge_case__html_non_lazy_builder']->embeddedAjaxResponseCommands),
    ]);

    $this->pass('Verifying BigPipe assets are present…', 'Debug');
    $this->assertFalse(empty($this->getDrupalSettings()), 'drupalSettings present.');
    $this->assertTrue(in_array('big_pipe/big_pipe', explode(',', $this->getDrupalSettings()['ajaxPageState']['libraries'])), 'BigPipe asset library is present.');
  }

  /**
   * Tests BigPipe-delivered HTML responses when JavaScript is disabled.
   *
   * Covers:
   * - \Drupal\big_pipe\EventSubscriber\HtmlResponseBigPipeSubscriber
   * - \Drupal\big_pipe\Render\BigPipe
   * - \Drupal\big_pipe\Render\BigPipe::sendNoJsPlaceholders()
   *
   * @see \Drupal\big_pipe\Tests\BigPipePlaceholderTestCases
   */
  public function testBigPipeNoJs() {
    $this->drupalLogin($this->rootUser);
    $this->assertSessionCookieExists(TRUE);
    $this->assertBigPipeNoJsCookieExists(FALSE);

    // By calling performMetaRefresh() here, we simulate JavaScript being
    // disabled, because as far as the BigPipe module is concerned, it is
    // enabled in the browser when the BigPipe no-JS cookie is set.
    // @see setUp()
    // @see performMetaRefresh()
    $this->performMetaRefresh();
    $this->assertBigPipeNoJsCookieExists(TRUE);

    $this->drupalGet(Url::fromRoute('big_pipe_test'));
    $this->assertBigPipeResponseHeadersPresent();

    $cases = $this->getTestCases();
    $this->assertBigPipeNoJsPlaceholders([
      $cases['edge_case__invalid_html']->bigPipeNoJsPlaceholder           => $cases['edge_case__invalid_html']->embeddedHtmlResponse,
      $cases['html_attribute_value']->bigPipeNoJsPlaceholder              => $cases['html_attribute_value']->embeddedHtmlResponse,
      $cases['html_attribute_value_subset']->bigPipeNoJsPlaceholder       => $cases['html_attribute_value_subset']->embeddedHtmlResponse,
      $cases['html']->bigPipeNoJsPlaceholder                              => $cases['html']->embeddedHtmlResponse,
      $cases['edge_case__html_non_lazy_builder']->bigPipeNoJsPlaceholder  => $cases['edge_case__html_non_lazy_builder']->embeddedHtmlResponse,
    ]);

    $this->pass('Verifying there are no BigPipe placeholders & replacements…', 'Debug');
    $this->assertEqual('<none>', $this->drupalGetHeader('BigPipe-Test-Placeholders'));
    $this->pass('Verifying BigPipe start/stop signals are absent…', 'Debug');
    $this->assertNoRaw(static::START_SIGNAL, 'BigPipe start signal absent.');
    $this->assertNoRaw(static::STOP_SIGNAL, 'BigPipe stop signal absent.');

    $this->pass('Verifying BigPipe assets are absent…', 'Debug');
    $this->assertFalse(empty($this->getDrupalSettings()), 'drupalSettings and BigPipe asset library absent.');
  }

  protected function assertBigPipeResponseHeadersPresent() {
    $this->pass('Verifying BigPipe response headers…', 'Debug');
    $this->assertTrue(FALSE !== strpos($this->drupalGetHeader('Cache-Control'), 'private'), 'Cache-Control header set to "private".');
    $this->assertEqual('no-store, content="BigPipe/1.0"', $this->drupalGetHeader('Surrogate-Control'));
    $this->assertEqual('no', $this->drupalGetHeader('X-Accel-Buffering'));
  }

  /**
   * Asserts expected BigPipe no-JS placeholders are present and replaced.
   *
   * @param array $expected_big_pipe_nojs_placeholders
   *   Keys: BigPipe no-JS placeholder markup. Values: expected replacement
   *   markup.
   */
  protected function assertBigPipeNoJsPlaceholders(array $expected_big_pipe_nojs_placeholders) {
    $this->pass('Verifying BigPipe no-JS placeholders & replacements…', 'Debug');
    $this->assertSetsEqual(array_keys($expected_big_pipe_nojs_placeholders), array_map('rawurldecode', explode(' ', $this->drupalGetHeader('BigPipe-Test-No-Js-Placeholders'))));
    foreach ($expected_big_pipe_nojs_placeholders as $big_pipe_nojs_placeholder => $expected_replacement) {
      $this->pass('Checking whether the replacement for the BigPipe no-JS placeholder "' . $big_pipe_nojs_placeholder . '" is present:');
      $this->assertRaw($expected_replacement);
    }
  }

  /**
   * Asserts expected BigPipe placeholders are present and replaced.
   *
   * @param array $expected_big_pipe_placeholders
   *   Keys: BigPipe placeholder IDs. Values: expected AJAX response.
   */
  protected function assertBigPipePlaceholders(array $expected_big_pipe_placeholders) {
    $this->pass('Verifying BigPipe placeholders & replacements…', 'Debug');
    $this->assertSetsEqual(array_keys($expected_big_pipe_placeholders), explode(' ', $this->drupalGetHeader('BigPipe-Test-Placeholders')));
    $placeholder_positions = [];
    $placeholder_replacement_positions = [];
    foreach ($expected_big_pipe_placeholders as $big_pipe_placeholder_id => $expected_ajax_response) {
      $this->pass('BigPipe placeholder: ' . $big_pipe_placeholder_id, 'Debug');
      // Verify expected placeholder.
      $expected_placeholder_html = '<div data-big-pipe-placeholder-id="' . $big_pipe_placeholder_id . '"></div>';
      $this->assertRaw($expected_placeholder_html, 'BigPipe placeholder for placeholder ID "' . $big_pipe_placeholder_id . '" found.');
      $pos = strpos($this->getRawContent(), $expected_placeholder_html);
      $placeholder_positions[$pos] = $big_pipe_placeholder_id;
      // Verify expected placeholder replacement.
      $result = $this->xpath('//script[@data-big-pipe-replacement-for-placeholder-with-id=:id]', [':id' => Html::decodeEntities($big_pipe_placeholder_id)]);
      $this->assertEqual($expected_ajax_response, trim((string) $result[0]));
      $expected_placeholder_replacement = '<script type="application/json" data-big-pipe-replacement-for-placeholder-with-id="' . $big_pipe_placeholder_id . '">';
      $this->assertRaw($expected_placeholder_replacement);
      $pos = strpos($this->getRawContent(), $expected_placeholder_replacement);
      $placeholder_replacement_positions[$pos] = $big_pipe_placeholder_id;
    }
    ksort($placeholder_positions, SORT_NUMERIC);
    $this->assertEqual(array_keys($expected_big_pipe_placeholders), array_values($placeholder_positions));
    $this->assertEqual(count($expected_big_pipe_placeholders), preg_match_all('/' . preg_quote('<div data-big-pipe-placeholder-id="', '/') . '/', $this->getRawContent()));
    $this->assertEqual(array_keys($expected_big_pipe_placeholders), array_values($placeholder_replacement_positions));
    $this->assertEqual(count($expected_big_pipe_placeholders), preg_match_all('/' . preg_quote('<script type="application/json" data-big-pipe-replacement-for-placeholder-with-id="', '/') . '/', $this->getRawContent()));

    $this->pass('Verifying BigPipe start/stop signals…', 'Debug');
    $this->assertRaw(static::START_SIGNAL, 'BigPipe start signal present.');
    $this->assertRaw(static::STOP_SIGNAL, 'BigPipe stop signal present.');
    $start_signal_position = strpos($this->getRawContent(), static::START_SIGNAL);
    $stop_signal_position = strpos($this->getRawContent(), static::STOP_SIGNAL);
    $this->assertTrue($start_signal_position < $stop_signal_position, 'BigPipe start signal appears before stop signal.');

    $this->pass('Verifying BigPipe placeholder replacements and start/stop signals were streamed in the correct order…', 'Debug');
    $expected_stream_order = array_keys($expected_big_pipe_placeholders);
    array_unshift($expected_stream_order, static::START_SIGNAL);
    array_push($expected_stream_order, static::STOP_SIGNAL);
    $actual_stream_order = $placeholder_replacement_positions + [
        $start_signal_position => static::START_SIGNAL,
        $stop_signal_position => static::STOP_SIGNAL,
      ];
    ksort($actual_stream_order, SORT_NUMERIC);
    $this->assertEqual($expected_stream_order, array_values($actual_stream_order));
  }

  protected function getTestCases() {
    // Ensure we can generate CSRF tokens for the current user's session.
    $session_data = $this->container->get('session_handler.write_safe')->read($this->cookies[$this->getSessionName()]['value']);
    $csrf_token_seed = unserialize(explode('_sf2_meta|', $session_data)[1])['s'];
    $this->container->get('session_manager.metadata_bag')->setCsrfTokenSeed($csrf_token_seed);

    return \Drupal\big_pipe\Tests\BigPipePlaceholderTestCases::cases($this->container, $this->rootUser);
  }

  /**
   * Asserts whether arrays A and B are equal, when treated as sets.
   */
  protected function assertSetsEqual(array $a, array $b) {
    $a_set = $a;
    sort($a_set);
    $b_set = $b;
    sort($b_set);
    return $this->assertEqual($a_set, $b_set);
  }

  /**
   * Asserts whether a BigPipe no-JS cookie exists or not.
   */
  protected function assertBigPipeNoJsCookieExists($expected) {
    $this->assertCookieExists('big_pipe_nojs', $expected, 'BigPipe no-JS');
  }

  /**
   * Asserts whether a session cookie exists or not.
   */
  protected function assertSessionCookieExists($expected) {
    $this->assertCookieExists($this->getSessionName(), $expected, 'Session');
  }

  /**
   * Asserts whether a cookie exists on the client or not.
   */
  protected function assertCookieExists($cookie_name, $expected, $cookie_label) {
    $non_deleted_cookies = array_filter($this->cookies, function ($item) { return $item['value'] !== 'deleted'; });
    $this->assertEqual($expected, isset($non_deleted_cookies[$cookie_name]), $expected ? "$cookie_label cookie exists." : "$cookie_label cookie does not exist.");
  }

  /**
   * Calls ::performMetaRefresh() and asserts the responses.
   */
  protected function assertBigPipeNoJsMetaRefreshRedirect() {
    $original_url = $this->url;
    $this->performMetaRefresh();

    $this->assertEqual($original_url, $this->url, 'Redirected back to the original location.');

    $headers = $this->drupalGetHeaders(TRUE);
    $this->assertEqual(2, count($headers), 'Two requests were made upon following the <meta> refresh, there are headers for two responses.');

    // First response: redirect.
    $this->assertEqual('HTTP/1.1 302 Found', $headers[0][':status'], 'The first response was a 302 (redirect).');
    $this->assertIdentical(0, strpos($headers[0]['set-cookie'], 'big_pipe_nojs=1'), 'The first response sets the big_pipe_nojs cookie.');
    $this->assertEqual($original_url, $headers[0]['location'], 'The first response redirected back to the original page.');
    $this->assertTrue(empty(array_diff(['cookies:big_pipe_nojs', 'session.exists'], explode(' ', $headers[0]['x-drupal-cache-contexts']))), 'The first response varies by the "cookies:big_pipe_nojs" and "session.exists" cache contexts.');
    $this->assertFalse(isset($headers[0]['surrogate-control']), 'The first response has no "Surrogate-Control" header.');

    // Second response: redirect followed.
    $this->assertEqual('HTTP/1.1 200 OK', $headers[1][':status'], 'The second response was a 200.');
    $this->assertTrue(empty(array_diff(['cookies:big_pipe_nojs', 'session.exists'], explode(' ', $headers[0]['x-drupal-cache-contexts']))), 'The first response varies by the "cookies:big_pipe_nojs" and "session.exists" cache contexts.');
    $this->assertEqual('no-store, content="BigPipe/1.0"', $headers[1]['surrogate-control'], 'The second response has a "Surrogate-Control" header.');

    $this->assertNoRaw('<noscript><meta http-equiv="Refresh" content="0; URL=', 'Once the BigPipe no-JS cookie is set, the <meta> refresh is absent: only one redirect ever happens.');
  }

}
